/* TraceCtl - Generic trace control loop
   Assumes trace facility configured by TrcInit.cmd

   Copyright (c) 2001-2020 Steven Levine and Associates, Inc.
   All rights reserved.

   This program is free software licensed under the terms of the GNU
   General Public License.  The GPL Software License can be found in
   gnugpl2.txt or at http://www.gnu.org/licenses/licenses.html#GPL

   2001-11-26 SHL Baseline
   2015-10-25 SHL Sync with templates
   2018-03-28 SHL Convert to Globals style
   2018-03-28 SHL Sync with templates
   2020-12-13 SHL Sync with templates
   2020-12-13 SHL Active tracefmt is already running
*/

signal on Error
signal on Failure name Error
signal on Halt
signal on NotReady name Error
signal on NoValue name Error
signal on Syntax name Error

gVersion = '0.3'

Globals = 'gCmdName gEnv gVersion'

call Initialize

Main:

  say
  say gCmdName 'starting.'

  TRCEXEPATH = value('TRCEXEPATH',,gEnv)

  /* Find boot drive */
  BootDrv = SysBootDrive()

  /* Must run from trace directory to avoid trap setting kernel tracepoints
     FIXME to know if still true
  */
  oldDir = directory()
  call Directory BootDrv'\os2\system\trace'

  /* Check trace initialized */
  signal off ERROR
  '@trace /q >nul'
  signal on ERROR
  if RC = 1055 then do
    say 'Trace not initialized'
    exit RC
  end

  /* Clear trace buffer */
  'trace /s /c'
  'trace /q'
  say

  req = ''
  opt = '/s'
  newOpt = '/r'
  pid = GetPID('tracefmt.exe')
  fStartTraceFmt = pid == ''

  do while req \= 'Q'

    if TRCEXEPATH \== '' then do
      /* List is comma separated */
      s = translate(TRCEXEPATH, ' ', ',')
      do i = 1 to words(s)
	exe = word(s, i)
	pid = GetPID(exe)
	if pid == '' then
	  call WarnMsgWarble 'Warning:' exe 'not running'
      end
    end

    if opt \= newOpt then do

      opt = newOpt
      'trace' opt

      if opt = '/r' then
	say 'Tracing resumed.'
      else do
	say 'Tracing suspended.'
	pid = GetPID('tracefmt.exe')
	fStartTraceFmt = pid == ''
      end
    end /* if newOpt */

    if fStartTraceFmt then do
      fStartTraceFmt = 0
      pid = GetPID('tracefmt.exe')
      if pid \= '' then do
	say 'Tracefmt already running - activating'
	'activate "OS/2 System Trace Formatter*" restore'
      end
      else do
	say 'Starting tracefmt'
	signal off Error		/* In case 457 */
	'start tracefmt'
	signal on Error
	call SysSleep 2
	pid = GetPID('tracefmt.exe')
	if pid == '' then
	  call WarnMsgWarble 'Warning: can not start tracefmt'
      end
    end

    say
    call charout ,'S)uspend R)esume O)ptions F)ormat Q)uit H)elp ? '
    req = translate(SysGetKey('NOECHO'))
    say
    select
    when req == 'H' then do
      say
      say ' S      suspend tracing'
      say ' R      resume tracing'
      say ' Enter  toggle trace options'
      say ' O      show trace options'
      say ' Q      turn off tracing and quit'
      say ' F      start trace formatter'
      say ' H      this message'
      say ' !      invoke shell'
    end
    when req == 'F' then
      fStartTraceFmt = 1
    when req == 'O' then do
      say
      'trace /q'
    end
    when req == 'R' then
      newOpt = '/r'
    when req == 'S' then
      newOpt = '/s'
    when req == x2c('0d') then do
      /* Enter Key - toggle state */
      newOpt = translate(opt, 'sr', 'rs')
    end
    when req == 'Q' then
      nop
    when req == '!' then do
      sShell = value('OS2_SHELL',,gEnv)
      say
      say 'Type exit to return'
      signal off Error
      '@'sShell
      signal on Error
    end
    otherwise
      nop				/* Ignore others */
    end /* select */

  end /* forever */

  /* Clean up */
  say
  'trace /c /s'
  'trace off'
  'trace off /p:all'
  say
  'trace /q'

  call directory oldDir

  exit

/* end main */

/*=== GetPID(exeName) Get PID for EXE ===*/

GetPID: procedure

  parse upper arg reqExe

  pidOut = ''

  '@pstat /c | rxqueue'

  matchLen = length(reqExe)

  /* Matches on first if multiple hits */

  do while queued() \= 0

    pull s
    /*
	      Parent
    Process   Process   Session   Process   Thread
      ID        ID        ID       Name       ID    Priority   Block ID   State
     00C9      0000       11      F:\IBMLAN\SERVICES\PEER.EXE    01      0300     FFFE0ACB   Block
   */
    if pidOut == '' & verify(substr(s, 2, 1), '0123456789ABCDEF', 'M') = 1
    then do
      /* Got process line */
      parse var s pid . . pstatExe .
      if pos('\', reqExe) = 0 then
	i = lastpos('\', pstatExe)	/* Match just exe name */
      else
	i = 0				/* Match full pathname */
      if substr(pstatExe, i + 1, matchLen) == reqExe then
	pidOut = pid
    end /* if */

  end /* while queued */

  if 0 then say 'GetPID('reqExe') returns' pidOut

  return pidOut

/* end GetPID */

/*=== Initialize() Initialize globals ===*/

Initialize: procedure expose (Globals)

  call SetCmdName
  call LoadRexxUtil
  gEnv = 'OS2ENVIRONMENT'
  return

/* end Initialize */

/*=== ScanArgsUsage(message) Report ScanArgsUsage Error... ===*/

ScanArgsUsage:
  parse arg msg
  say msg
  say 'Usage:' gCmdName
  exit 255

/* end ScanArgsUsage */

/*=== ScanArgsHelp() Display help ===*/

ScanArgsHelp:

  say
  say 'Usage:' gCmdName '[-h]'
  say
  say ' -h  This message'

  exit 255

/* end ScanArgsHelp */

/*==============================================================================*/
/*=== SkelRexxFunc standards - Delete unused - Move modified above this mark ===*/
/*==============================================================================*/

/*=== WarnMsgWarble(message) Write multi-line warning message to STDERR and warble ===*/

WarnMsgWarble: procedure
  do i = 1 to arg()
    msg = arg(i)
    call lineout 'STDERR', msg
  end
  do 10;
    f=random(262,1047)
    d=random(100,200)
    call beep f,d
  end
  return

/* end WarnMsgWarble */

/*==========================================================================*/
/*=== SkelRexx standards - Delete unused - Move modified above this mark ===*/
/*==========================================================================*/

/*=== Die([message,...]) Write multi-line message to STDERR and die ===*/

Die:
  call lineout 'STDERR', ''
  do i = 1 to arg()
    call lineout 'STDERR', arg(i)
  end
  call lineout 'STDERR', gCmdName 'aborting at line' SIGL || '.'
  call beep 200, 300
  call SysSleep 2
  exit 254

/* end Die */

/*=== Error() Set gErrCondition; report to STDOUT; trace and exit or return if called ===*/

Error:
  say
  parse source . . cmd
  gErrCondition = condition('C')
  say gErrCondition 'signaled at line' SIGL 'of' cmd || '.'
  if condition('D') \= '' then
    say 'REXX reason =' condition('D') || '.'
  if gErrCondition == 'SYNTAX' & symbol('RC') == 'VAR' then
    say 'REXX error =' RC '-' errortext(RC) || '.'
  else if symbol('RC') == 'VAR' then
    say 'RC =' RC || '.'
  say 'Source =' sourceline(SIGL)

  if condition('I') \== 'CALL' | gErrCondition == 'NOVALUE' | gErrCondition == 'SYNTAX' then do
    trace '?A'
    say 'Enter REXX commands to debug failure.  Press enter to exit script.'
    nop
    if symbol('RC') \== 'VAR' then
      RC = 255
    exit RC
  end

  return

/* end Error */

/*=== Halt() Report HALT condition to STDOUT and exit ===*/

Halt:
  say
  parse source . . cmd
  say condition('C') 'signaled at' cmd 'line' SIGL || '.'
  say 'Source =' sourceline(SIGL)
  say 'Sleeping for 2 seconds...'
  call SysSleep 2
  exit 253

/* end Halt */

/*=== LoadRexxUtil() Load RexxUtil functions ===*/

LoadRexxUtil:
  if RxFuncQuery('SysLoadFuncs') then do
    call RxFuncAdd 'SysLoadFuncs', 'REXXUTIL', 'SysLoadFuncs'
    if RESULT then
      call Die 'Cannot load SysLoadFuncs.'
    call SysLoadFuncs
  end
  return

/* end LoadRexxUtil */

/*=== SetCmdName() Set gCmdName to short script name ===*/

SetCmdName: procedure expose (Globals)
  parse source . . cmd
  cmd = filespec('N', cmd)		/* Chop path */
  c = lastpos('.', cmd)
  if c > 1 then
    cmd = left(cmd, c - 1)		/* Chop extension */
  gCmdName = translate(cmd, xrange('a', 'z'), xrange('A', 'Z'))	/* Lowercase */
  return

/* end SetCmdName */

/* The end */
