/* TrcInit - Initialize OS/2 Trace using configuration file

   Copyright (c) 2001-2020 Steven Levine and Associates, Inc.
   All rights reserved.

   This program is free software licensed under the terms of the GNU
   General Public License.  The GPL Software License can be found in
   gnugpl2.txt or at http://www.gnu.org/licenses/licenses.html#GPL

   2001-01-11 SHL Baseline
   2015-10-25 SHL Shut down trace before dieing
   2018-03-28 SHL Convert to Globals style
   2018-03-28 SHL Sync with templates
   2018-03-28 SHL Support OPTIONS keyword
   2018-03-29 SHL Avoid spurious buffer size change errors
   2020-12-13 SHL Sync with templates
   2020-12-13 SHL Strip spaces from command line

   Config file options
     CMD = tracecmd		trace setup commands
     DIEONERROR = yes/no	die on errors
     EXE = exe			exe to trace
     FINDPID = yes/no		find PID for exe
     OPTIONS			override default options (i.e. /b:512)
     PID = pid			pid to trace (overrides exe)

   Notes: tracepoints are set in decimal and displayed in hex
	  pids are set in hex and displayed in hex
	  krnlrfs tracepoints 281 and 283 will trap on 14.93c to 14.100c

*/

/* signal on Error */
/* signal on Failure name Error */
signal on Halt
signal on NotReady name Error
signal on NoValue name Error
signal on Syntax name Error

gVersion = '0.8'

Globals = 'gCmdName gEnv gTrc. gVersion'

call Initialize

Main:

  say
  say gCmdName 'starting.'

  '@setlocal'

  parse arg cmdLine
  cmdLine = strip(cmdLine)

  if cmdLine = '-h' | cmdLine = '-?' then
    call ScanArgsHelp
  if left(cmdLine, 1) = '-' then
    call ScanArgsUsage

  call ReadCfg cmdLine

  drop cmdLine

  /* Build comma sep EXE list, find PIDs if requested */
  exe = ''
  fail = 0
  do iExe = 1 to gTrc.!EXE.0
    if iExe > 1 then
      exe = exe','
    exe = exe || gTrc.!EXE.iExe
    if gTrc.!FindPID then do
      pid = EXE2PID(gTrc.!EXE.iExe)
      if pid == '' then do
	call WarnMsgWarble 'Warning:' gTrc.!EXE.iExe 'not running'
	fail = 1
      end
      else do
	iPid = gTrc.!PID.0 + 1
	gTrc.!PID.iPid = pid
	gTrc.!PID.0 = iPid
      end
    end
  end

  if fail then
    call Die 'Please check configuration file'	/* Time to die */

  /* Pass EXE names to others as comma sep list */
  call value 'TRCEXEPATH', exe, gEnv

  /* Build PID list */
  pid = ''
  do iPid = 1 to gTrc.!PID.0
    if pid \== '' then
      pid = pid','
    pid = pid || gTrc.!pid.iPID
  end

  /* Find boot drive */
  BootDrv = SysBootDrive()

  /* Find trcctl.cmd in case in current dir and not in PATH */
  trcCtlCmd = SysSearchPath('PATH', 'trcctl.cmd')

  if trcCtlCmd == '' then
    call Die 'trcctl.cmd not found'

  /* Must run from here to avoid trap setting kernel tracepoints - fixme still true? */
  oldDir = directory()
  call Directory BootDrv'\os2\system\trace'

  /* Ensure trace initialized */
  opts = gTrc.!Options
  haveBuf = pos('/B:', translate(opts)) > 0
  '@trace /q >nul'
  if RC = 1055 then do
    if \ haveBuf then do
      opts = strip('/b:512' opts)
      haveBuf = 1
    end
  end
  if haveBuf & RC = 0 then do
    'trace off /b'
    if RC \= 0 then
      call Die 'Trace initialzation failed with error' RC
  end
  say 'Initializing trace with options' opts
  'trace on' opts
  if haveBuf & RC = 1055 then do
    /* 1055 may be bogus */
    '@trace /q >nul'
  end
  if RC \= 0 then
    call Die 'Trace initialzation failed with error' RC

  'trace off'
  'trace /c /s'

  do i = 1 to gTrc.!Cmd.0
    if 0 then say gTrc.!Cmd.i
    gTrc.!Cmd.i
    if RC \= 0 then do
      call WarnMsgWarble 'Warning: can not set tracepoint(s) - trace reported error' RC'.'
      if gTrc.!DieOnError then do
	/* Clean up */
	'trace off'
	'trace off /p:all'
	call Die 'Die on error requested by configuration.'
      end
    end
  end

  call Directory oldDir

  /* Trace specific process, if requested */
  if pid \= '' then
    'trace on /p:'pid
  else if exe \= '' then
    'trace on /n:'exe
  else
    'trace on /p:all'			/* Just in case */

  interpret 'call '''trcCtlCmd''''

  exit

/* end main */

/*=== Die(message) Beep and write message to STDERR and die ===*/

Die: procedure
  parse arg msg
  call Warn1 msg
  exit

/* end Die */

/*=== EXE2PID(exeName) Get PID for EXE, return hex PID or empty string ===*/

EXE2PID: procedure

  parse upper arg exeNameIn

  hexPIDOut = ''

  matchCnt = length(exeNameIn)

  '@pstat /c | rxqueue'

  /* Match on first if multiple hits */

  do while queued() \= 0

    pull s

    if hexPIDOut \= '' then iterate	/* skip rest */

    /*
	      Parent
    Process   Process   Session   Process   Thread
      ID        ID        ID       Name       ID    Priority   Block ID   State
     00C9      0000       11      F:\IBMLAN\SERVICES\PEER.EXE    01      0300     FFFE0ACB   Block
   */

    if verify(substr(s, 2, 1), '0123456789ABCDEF', 'M') = 1
    then do
      /* Got process line */
      parse var s hexPID . . pathName .
      if pos('\', exeNameIn) = 0 then
	i = lastpos('\', pathName)	/* Match just exe name */
      else
	i = 0				/* Match full pathname */
      if substr(pathName, i + 1, matchCnt) == exeNameIn then
	hexPIDOut = hexPID
    end

  end /* while queued */

  return hexPIDOut

/* end EXE2PID */

/*=== PID2EXE(hexPID) Get EXE for PID, return pathname or empty string ===*/

PID2EXE: procedure

  parse arg hexPIDIn

  hexPIDIn = right('0000' || translate(hexPIDIn), 4)

  '@pstat /c | rxqueue'

  /* Match on first if multiple hits */
  exe = ''

  do while queued() \= 0

    pull s

    if exe \= '' then iterate		/* skip rest */

    /*
	      Parent
    Process   Process   Session   Process   Thread
      ID        ID        ID       Name       ID    Priority   Block ID   State
     00C9      0000       11      F:\IBMLAN\SERVICES\PEER.EXE    01      0300     FFFE0ACB   Block
   */

    if verify(substr(s, 2, 1), '0123456789ABCDEF', 'M') = 1
    then do
      /* Got process line */
      parse var s hexPID . . sPathName .
      if hexPIDIn = hexPID then
	exe = sPathName
    end

  end /* while queued */

  return exe

/* end PID2EXE */

/*=== ReadCfg(cfgFile) Read settings from trcinit.cfg ===*/

ReadCfg: procedure expose (Globals)

  parse arg cfgFile

  if cfgFile = '' then
    cfgFile = gCmdName'.cfg'		/* Default */

  /* Preset */
  gTrc.!Cmd.0 = 0
  gTrc.!DieOnError = 1
  gTrc.!EXE.0 = 0
  gTrc.!FindPID = 0
  gTrc.!PID.0 = 0
  gTrc.!Options = ''

  /* Scan and parse */

  say 'Reading' cfgFile

  s = stream(cfgFile, 'C', 'QUERY EXISTS')

  if s == '' then do
    i = lastpos('.', cfgFile)
    j = lastpos('\', cfgFile)
    if i <= j then do
      s = stream(cfgFile'.cfg', 'C', 'QUERY EXISTS')
      if s \== '' then
	cfgFile = s
    end
    if s == '' then
      call ScanArgsUsage cfgFile 'does not exist.'
  end

  fail = 0
  warnmsg = ''
  failmsg = ''
  linenum = 0

  call stream cfgFile, 'C', 'OPEN READ'

  do while lines(cfgFile) \= 0

    line = linein(cfgFile)
    line = strip(line)
    linenum = linenum + 1

    if line = '' then iterate
    if left(line, 1) = ';' then iterate	/* Comment line */

    parse var line req '=' opt
    req = strip(req)
    opt = strip(opt)

    uopt = translate(opt)
    select
    when abbrev('YES', uopt) \= 0 then yesno = 1
    when uopt = 1 then yesno = 1
      when abbrev('NO', uopt) \= 0 then yesno = 0
    when uopt = 0 then yesno = 0
    otherwise yesno = ''
    end

    ureq = translate(req)
    select
    when abbrev('CMD', ureq) \= 0 then do
      if opt = '' then
	failmsg = 'Expected command line for' req
      else do
	i = gTrc.!Cmd.0 + 1
	gTrc.!Cmd.i = opt
	gTrc.!Cmd.0 = i
      end
    end
    when abbrev('DIEONERROR', ureq) \= 0 then do
      if yesno = '' then
	failmsg 'Expected yes/no value for' req
      else
	gTrc.!DieOnError = yesno
    end
    when abbrev('EXE', ureq) \= 0 then do
      if opt = '' then
	failmsg = 'Expected exe name for' req
      else do
	i = lastpos('.', opt)
	j = lastpos('\', opt)
	if i <= j then
	  opt = opt'.exe'
	exe = stream(opt, 'C', 'QUERY EXISTS')
	if exe == '' then
	  exe = SysSearchPath('PATH', opt)
	if exe == '' then do
	  warnmsg = opt 'not found by name or by PATH'
	  exe = opt
	end
	/* Always add to list */
	i = gTrc.!EXE.0 + 1
	gTrc.!EXE.i = exe
	gTrc.!EXE.0 = i
      end
    end
    when abbrev('FINDPID', ureq) \= 0 then do
      if yesno = '' then
	failmsg = 'Expected yes/no value for' req
      else
	gTrc.!FindPID = yesno
    end
    when abbrev('PID', ureq) \= 0 then do
      if datatype(opt, 'X') = 0 then
	failmsg = 'Expected hexadecimal PID number for' req
      else do
	exeNameOut = PID2EXE(opt)
	if exeNameOut = '' then
	  failmsg = 'PID' opt 'not running'
	else do
	  say 'PID' opt 'is' exeNameOut
	  i = gTrc.!EXE.0 + 1
	  gTrc.!EXE.i = exeNameOut
	  gTrc.!PID.i = opt
	  gTrc.!EXE.0 = i
	end
      end
    end
    when abbrev('OPTIONS', ureq) \= 0 then do
      if gTrc.!Options \== '' then
	failmsg = 'Options already set to' gTrc.!Options
      else do
	gTrc.!Options = opt
      end
    end
    otherwise
      failmsg = 'Request "' || req || '" unexpected at line' linenum 'of' cfgFile
    end

    if failmsg \= '' then do
      call Warn1 failmsg
      failmsg = ''
      fail = 1
    end

    if warnmsg \= '' then do
      call Warn1 warnmsg
      warnmsg = ''
      if gTrc.!DieOnError then
	fail = 1
    end

  end /* while lines */

  call stream cfgFile, 'C', 'CLOSE'

  if gTrc.!PID.0 > 0 & gTrc.!EXE.0 > 0 & \ gTrc.!FindPID then do
    call Warn1 'PID not allowed with EXE'
    fail = 1
  end

  if fail then
    call Die 'Please check configuration file'	/* Time to die */

  return

/* end ReadCfg */

/*=== Initialize: Initialize globals ===*/

Initialize: procedure expose (Globals)
  call LoadRexxUtil
  call SetCmdName
  gEnv = 'OS2ENVIRONMENT'		/* Used to find TRCEXEPATH etc. */
  return

/* end Initialize */

/*=== ScanArgsHelp() Display usage help ===*/

ScanArgsHelp:

  say
  say 'Usage:' gCmdName '[cfgFile]'
  say
  say ' cfgFile  configuration file, defaults to' gCmdName'.cfg'

  exit 255

/* end ScanArgsHelp */

/*=== ScanArgsUsage(message) Report usage error ===*/

ScanArgsUsage:
  parse arg msg
  say
  if msg \= '' then
    say msg
  say 'Usage:' gCmdName '[cfgFile]'
  exit 255

/* end ScanArgsUsage */

/*==============================================================================*/
/*=== SkelRexxFunc standards - Delete unused - Move modified above this mark ===*/
/*==============================================================================*/

/*=== Warn1(message) Beep and write message to STDERR ===*/

Warn1: procedure
  parse arg msg
  call lineout 'STDERR', msg
  call Beep 400, 300
  return

/* end Warn1 */

/*=== WarnMsgWarble(message) Write multi-line warning message to STDERR and warble ===*/

WarnMsgWarble: procedure
  do i = 1 to arg()
    msg = arg(i)
    call lineout 'STDERR', msg
  end
  do 10;
    f=random(262,1047)
    d=random(100,200)
    call beep f,d
  end
  return

/* end WarnMsgWarble */

/*==========================================================================*/
/*=== SkelRexx standards - Delete unused - Move modified above this mark ===*/
/*==========================================================================*/

/*=== Die([message,...]) Write multi-line message to STDERR and die ===*/

Die:
  call lineout 'STDERR', ''
  do i = 1 to arg()
    call lineout 'STDERR', arg(i)
  end
  call lineout 'STDERR', gCmdName 'aborting at line' SIGL || '.'
  call beep 200, 300
  call SysSleep 2
  exit 254

/* end Die */

/*=== Error() Set gErrCondition; report to STDOUT; trace and exit or return if called ===*/

Error:
  say
  parse source . . cmd
  gErrCondition = condition('C')
  say gErrCondition 'signaled at line' SIGL 'of' cmd || '.'
  if condition('D') \= '' then
    say 'REXX reason =' condition('D') || '.'
  if gErrCondition == 'SYNTAX' & symbol('RC') == 'VAR' then
    say 'REXX error =' RC '-' errortext(RC) || '.'
  else if symbol('RC') == 'VAR' then
    say 'RC =' RC || '.'
  say 'Source =' sourceline(SIGL)

  if condition('I') \== 'CALL' | gErrCondition == 'NOVALUE' | gErrCondition == 'SYNTAX' then do
    trace '?A'
    say 'Enter REXX commands to debug failure.  Press enter to exit script.'
    nop
    if symbol('RC') \== 'VAR' then
      RC = 255
    exit RC
  end

  return

/* end Error */

/*=== Halt() Report HALT condition to STDOUT and exit ===*/

Halt:
  say
  parse source . . cmd
  say condition('C') 'signaled at' cmd 'line' SIGL || '.'
  say 'Source =' sourceline(SIGL)
  say 'Sleeping for 2 seconds...'
  call SysSleep 2
  exit 253

/* end Halt */

/*=== LoadRexxUtil() Load RexxUtil functions ===*/

LoadRexxUtil:
  if RxFuncQuery('SysLoadFuncs') then do
    call RxFuncAdd 'SysLoadFuncs', 'REXXUTIL', 'SysLoadFuncs'
    if RESULT then
      call Die 'Cannot load SysLoadFuncs.'
    call SysLoadFuncs
  end
  return

/* end LoadRexxUtil */

/*=== SetCmdName() Set gCmdName to short script name ===*/

SetCmdName: procedure expose (Globals)
  parse source . . cmd
  cmd = filespec('N', cmd)		/* Chop path */
  c = lastpos('.', cmd)
  if c > 1 then
    cmd = left(cmd, c - 1)		/* Chop extension */
  gCmdName = translate(cmd, xrange('a', 'z'), xrange('A', 'Z'))	/* Lowercase */
  return

/* end SetCmdName */

/* The end */
